/*
 * Copyright (c) 2022 Jonathan Schleifer <js@nil.im>
 *
 * All rights reserved.
 *
 * This file is part of Obj3DEngine. It may be distributed under the terms of
 * the Q Public License 1.0, which can be found in the file LICENSE.QPL
 * included in the packaging of this file.
 *
 * Alternatively, it may be distributed under the terms of the GNU General
 * Public License, either version 2 or 3, which can be found in the file
 * LICENSE.GPLv2 or LICENSE.GPLv3 respectively included in the packaging of
 * this file.
 */

#import "O3DGlide3Renderer.h"

static OFSize
screenResolutionToSize(GrScreenResolution_t resolution)
{
	switch (resolution) {
	case GR_RESOLUTION_320x200:
		return OFMakeSize(320, 200);
	case GR_RESOLUTION_320x240:
		return OFMakeSize(320, 240);
	case GR_RESOLUTION_400x256:
		return OFMakeSize(400, 256);
	case GR_RESOLUTION_512x384:
		return OFMakeSize(512, 384);
	case GR_RESOLUTION_640x200:
		return OFMakeSize(640, 200);
	case GR_RESOLUTION_640x350:
		return OFMakeSize(640, 350);
	case GR_RESOLUTION_640x400:
		return OFMakeSize(640, 400);
	case GR_RESOLUTION_640x480:
		return OFMakeSize(640, 480);
	case GR_RESOLUTION_800x600:
		return OFMakeSize(800, 600);
	case GR_RESOLUTION_960x720:
		return OFMakeSize(960, 720);
	case GR_RESOLUTION_856x480:
		return OFMakeSize(856, 480);
	case GR_RESOLUTION_512x256:
		return OFMakeSize(512, 256);
	case GR_RESOLUTION_1024x768:
		return OFMakeSize(1024, 768);
	case GR_RESOLUTION_1280x1024:
		return OFMakeSize(1280, 1024);
	case GR_RESOLUTION_1600x1200:
		return OFMakeSize(1600, 1200);
	case GR_RESOLUTION_400x300:
		return OFMakeSize(400, 300);
	case GR_RESOLUTION_1152x864:
		return OFMakeSize(1152, 864);
	case GR_RESOLUTION_1280x960:
		return OFMakeSize(1280, 960);
	case GR_RESOLUTION_1600x1024:
		return OFMakeSize(1600, 1024);
	case GR_RESOLUTION_1792x1344:
		return OFMakeSize(1792, 1344);
	case GR_RESOLUTION_1856x1392:
		return OFMakeSize(1856, 1392);
	case GR_RESOLUTION_1920x1440:
		return OFMakeSize(1920, 1440);
	case GR_RESOLUTION_2048x1536:
		return OFMakeSize(2048, 1536);
	case GR_RESOLUTION_2048x2048:
		return OFMakeSize(2048, 2048);
	default:
		return OFMakeSize(0, 0);
	}
}

static GrScreenResolution_t
sizeToScreenResolution(OFSize size)
{
	if (size.width == 320 && size.height == 200)
		return GR_RESOLUTION_320x200;
	if (size.width == 320 && size.height == 240)
		return GR_RESOLUTION_320x240;
	if (size.width == 400 && size.height == 256)
		return GR_RESOLUTION_400x256;
	if (size.width == 512 && size.height == 384)
		return GR_RESOLUTION_512x384;
	if (size.width == 640 && size.height == 200)
		return GR_RESOLUTION_640x200;
	if (size.width == 640 && size.height == 350)
		return GR_RESOLUTION_640x350;
	if (size.width == 640 && size.height == 400)
		return GR_RESOLUTION_640x400;
	if (size.width == 640 && size.height == 480)
		return GR_RESOLUTION_640x480;
	if (size.width == 800 && size.height == 600)
		return GR_RESOLUTION_800x600;
	if (size.width == 960 && size.height == 720)
		return GR_RESOLUTION_960x720;
	if (size.width == 856 && size.height == 480)
		return GR_RESOLUTION_856x480;
	if (size.width == 512 && size.height == 256)
		return GR_RESOLUTION_512x256;
	if (size.width == 1024 && size.height == 768)
		return GR_RESOLUTION_1024x768;
	if (size.width == 1280 && size.height == 1024)
		return GR_RESOLUTION_1280x1024;
	if (size.width == 1600 && size.height == 1200)
		return GR_RESOLUTION_1600x1200;
	if (size.width == 400 && size.height == 300)
		return GR_RESOLUTION_400x300;
	if (size.width == 1152 && size.height == 864)
		return GR_RESOLUTION_1152x864;
	if (size.width == 1280 && size.height == 960)
		return GR_RESOLUTION_1280x960;
	if (size.width == 1600 && size.height == 1024)
		return GR_RESOLUTION_1600x1024;
	if (size.width == 1792 && size.height == 1344)
		return GR_RESOLUTION_1792x1344;
	if (size.width == 1856 && size.height == 1392)
		return GR_RESOLUTION_1856x1392;
	if (size.width == 1920 && size.height == 1440)
		return GR_RESOLUTION_1920x1440;
	if (size.width == 2048 && size.height == 1536)
		return GR_RESOLUTION_2048x1536;
	if (size.width == 2048 && size.height == 2048)
		return GR_RESOLUTION_2048x2048;

	return GR_RESOLUTION_NONE;
}

static float
screenRefreshToFloat(GrScreenRefresh_t refresh)
{
	switch (refresh) {
	case GR_REFRESH_60Hz:
		return 60;
	case GR_REFRESH_70Hz:
		return 70;
	case GR_REFRESH_72Hz:
		return 72;
	case GR_REFRESH_75Hz:
		return 75;
	case GR_REFRESH_80Hz:
		return 80;
	case GR_REFRESH_90Hz:
		return 90;
	case GR_REFRESH_100Hz:
		return 100;
	case GR_REFRESH_85Hz:
		return 85;
	case GR_REFRESH_120Hz:
		return 120;
	default:
		return 0;
	}
}

static GrScreenRefresh_t
floatToScreenRefresh(float refresh)
{
	if (refresh == 60)
		return GR_REFRESH_60Hz;
	if (refresh == 70)
		return GR_REFRESH_70Hz;
	if (refresh == 72)
		return GR_REFRESH_72Hz;
	if (refresh == 75)
		return GR_REFRESH_75Hz;
	if (refresh == 80)
		return GR_REFRESH_80Hz;
	if (refresh == 90)
		return GR_REFRESH_90Hz;
	if (refresh == 100)
		return GR_REFRESH_100Hz;
	if (refresh == 85)
		return GR_REFRESH_85Hz;
	if (refresh == 120)
		return GR_REFRESH_120Hz;

	return GR_REFRESH_NONE;
}

@implementation O3DGlide3Renderer
- (instancetype)initWithOptions:
    (OFDictionary OF_GENERIC(OFString *, id) *)options
{
	self = [super init];

	@try {
		grGlideInit();

		FxI32 numDevices;
		if (grGet(GR_NUM_BOARDS, sizeof(numDevices), &numDevices) == 0)
			@throw [OFInitializationFailedException
			    exceptionWithClass: self.class];

		long long deviceIndex = [[options objectForKey:
		    O3DRendererDeviceIndex] longLongValue];
		if (deviceIndex > numDevices)
			@throw [OFInvalidArgumentException exception];

		grSstSelect((int)deviceIndex);
	} @catch (id e) {
		[self release];
		@throw e;
	}

	return self;
}

- (void)dealloc
{
	if (_context != 0)
		grSstWinClose(_context);

	grGlideShutdown();

	[super dealloc];
}

- (OFArray OF_GENERIC(OFPair OF_GENERIC(OFValue *, OFNumber *) *) *)
    availableResolutions
{
	OFMutableArray *ret = [OFMutableArray array];
	GrResolution query = {
		.resolution = GR_QUERY_ANY,
		.refresh = GR_QUERY_ANY,
		.numColorBuffers = 2,
		.numAuxBuffers = 1
	};
	size_t size = grQueryResolutions(&query, NULL);

	GrResolution *resolutions = OFAllocMemory(1, size);
	@try {
		size = grQueryResolutions(&query, resolutions);

		void *pool = objc_autoreleasePoolPush();
		for (size_t i = 0; i < size / sizeof(*resolutions); i++) {
			OFSize resolutionSize =
			    screenResolutionToSize(resolutions[i].resolution);
			float refreshFloat =
			    screenRefreshToFloat(resolutions[i].refresh);

			if (resolutionSize.width == 0 ||
			    resolutionSize.height == 0)
				continue;
			if (refreshFloat == 0)
				continue;

			OFValue *resolution =
			    [OFValue valueWithSize: resolutionSize];
			OFNumber *refresh =
			    [OFNumber numberWithFloat: refreshFloat];
			[ret addObject: [OFPair pairWithFirstObject: resolution
						       secondObject: refresh]];
		}
		objc_autoreleasePoolPop(pool);
	} @finally {
		OFFreeMemory(resolutions);
	}

	[ret sortUsingFunction: O3DCompareResolution
		       context: NULL
		       options: OFArraySortDescending];

	[ret makeImmutable];

	return ret;
}

- (void)createWithResolution: (O3DResolution)resolution
{
	GrScreenResolution_t screenResolution =
	    sizeToScreenResolution([resolution.firstObject sizeValue]);
	GrScreenRefresh_t screenRefresh =
	    floatToScreenRefresh([resolution.secondObject floatValue]);

	if (screenResolution == GR_RESOLUTION_NONE ||
	    screenRefresh == GR_REFRESH_NONE)
		@throw [OFInvalidArgumentException exception];

	if ((_context = grSstWinOpen(0, screenRefresh, screenRefresh,
	    GR_COLORFORMAT_RGBA, GR_ORIGIN_LOWER_LEFT, 2, 1)) == 0)
		@throw [OFInitializationFailedException
		    exceptionWithClass: self.class];
}
@end
